#!/usr/bin/env python
# crate_anon/tools/estimate_mysql_memory_usage.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Script to check the memory usage (approximately) of a running MySQL instance.

From: https://dev.mysql.com/doc/refman/5.0/en/memory-use.html
- However, innodb_additional_mem_pool_size deprecated in 5.6.3 and removed in
  5.7.4; http://dev.mysql.com/doc/refman/5.7/en/innodb-parameters.html
"""


import argparse
import subprocess

import prettytable


MYSQL_DEFAULT_PORT = 3306
MYSQL_DEFAULT_USER = 'root'
UNITS_MB = 'Mb'


def get_mysql_vars(mysql, host, port, user):
    cmdargs = [
        mysql,
        "-h", host,
        "-P", str(port),
        "-e", "SHOW VARIABLES; SHOW STATUS",
        "-u", user,
        "-p"  # prompt for password
    ]
    print("Connecting to MySQL with user: {}".format(user))
    print(cmdargs)
    process = subprocess.Popen(cmdargs, stdout=subprocess.PIPE)
    out, err = process.communicate()
    lines = out.decode("utf8").splitlines()
    mysqlvars = {}
    for line in lines:
        var, val = line.split("\t")
        mysqlvars[var] = val
    return mysqlvars


def val_mb(valstr):
    try:
        return "{:.3f}".format(int(valstr) / (1024 * 1024))
    except (TypeError, ValueError):
        return '?'


def val_int(valstr):
    return str(valstr) + " " * 4


def add_var_mb(table, vardict, varname):
    valstr = vardict.get(varname, None)
    table.add_row([varname, val_mb(valstr), UNITS_MB])


def add_blank_row(table):
    table.add_row([''] * 3)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--mysql", default="mysql",
        help="MySQL program (default=mysql)")
    parser.add_argument(
        "--host", default="127.0.0.1",
        help="MySQL server/host (prefer '127.0.0.1' to 'localhost')")
    parser.add_argument(
        "--port", type=int, default=MYSQL_DEFAULT_PORT,
        help="MySQL port (default={})".format(MYSQL_DEFAULT_PORT))
    parser.add_argument(
        "--user", default=MYSQL_DEFAULT_USER,
        help="MySQL user (default={})".format(MYSQL_DEFAULT_USER))
    args = parser.parse_args()

    vardict = get_mysql_vars(
        mysql=args.mysql,
        host=args.host,
        port=args.port,
        user=args.user,
    )
    max_conn = int(vardict["max_connections"])
    max_used_conn = int(vardict["Max_used_connections"])
    base_mem = (
        int(vardict["key_buffer_size"]) +
        int(vardict["query_cache_size"]) +
        int(vardict["innodb_buffer_pool_size"]) +
        # int(vardict["innodb_additional_mem_pool_size"]) +
        int(vardict["innodb_log_buffer_size"])
    )
    mem_per_conn = (
        int(vardict["read_buffer_size"]) +
        int(vardict["read_rnd_buffer_size"]) +
        int(vardict["sort_buffer_size"]) +
        int(vardict["join_buffer_size"]) +
        int(vardict["binlog_cache_size"]) +
        int(vardict["thread_stack"]) +
        int(vardict["tmp_table_size"])
    )
    mem_total_min = base_mem + mem_per_conn * max_used_conn
    mem_total_max = base_mem + mem_per_conn * max_conn

    table = prettytable.PrettyTable(["Variable", "Value", "Units"])
    table.align["Variable"] = "l"
    table.align["Value"] = "r"
    table.align["Units"] = "l"

    add_var_mb(table, vardict, "key_buffer_size")
    add_var_mb(table, vardict, "query_cache_size")
    add_var_mb(table, vardict, "innodb_buffer_pool_size")
    # print_var_mb(table, vardict, "innodb_additional_mem_pool_size")
    add_var_mb(table, vardict, "innodb_log_buffer_size")
    add_blank_row(table)
    table.add_row(["BASE MEMORY", val_mb(base_mem), UNITS_MB])
    add_blank_row(table)
    add_var_mb(table, vardict, "sort_buffer_size")
    add_var_mb(table, vardict, "read_buffer_size")
    add_var_mb(table, vardict, "read_rnd_buffer_size")
    add_var_mb(table, vardict, "join_buffer_size")
    add_var_mb(table, vardict, "thread_stack")
    add_var_mb(table, vardict, "binlog_cache_size")
    add_var_mb(table, vardict, "tmp_table_size")
    add_blank_row(table)
    table.add_row(["MEMORY PER CONNECTION", val_mb(mem_per_conn), UNITS_MB])
    add_blank_row(table)
    table.add_row(["Max_used_connections", val_int(max_used_conn), ''])
    table.add_row(["max_connections", val_int(max_conn), ''])
    add_blank_row(table)
    table.add_row(["TOTAL (MIN)", val_mb(mem_total_min), UNITS_MB])
    table.add_row(["TOTAL (MAX)", val_mb(mem_total_max), UNITS_MB])

    print(table.get_string())


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/crateweb/research/html_functions.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
import re
import textwrap
import typing.re
from typing import Any, Dict, Iterable, List, Optional

from django.contrib.staticfiles.templatetags.staticfiles import static
from django.utils.html import escape
# from django.template import loader
from django.template.defaultfilters import linebreaksbr
from pygments import highlight
from pygments.lexers.sql import SqlLexer
from pygments.formatters.html import HtmlFormatter
import sqlparse

from crate_anon.crateweb.extra.django_cache_fn import django_cache_function

log = logging.getLogger(__name__)


N_CSS_HIGHLIGHT_CLASSES = 3  # named highlight0, highlight1, ... highlight<n-1>
REGEX_METACHARS = ["\\", "^", "$", ".",
                   "|", "?", "*", "+",
                   "(", ")", "[", "{"]
# http://www.regular-expressions.info/characters.html
# Start with \, for replacement.


# =============================================================================
# Collapsible div, etc.
# =============================================================================

def visibility_button(tag: str, small: bool = True,
                      title_html: str = '', as_span: bool = False,
                      as_visibility: bool = True) -> str:
    eltype = "span" if as_span else "div"
    togglefunc = "toggleVisible" if as_visibility else "toggleCollapsed"
    return """
<{eltype} class="expandcollapse" onclick="{togglefunc}('collapsible_{tag}', 'collapse_img_{tag}');">
    <img class="plusminus_image" id="collapse_img_{tag}" alt="" src="{img}">
    {title_html}
</{eltype}>
    """.format(  # noqa
        eltype=eltype,
        togglefunc=togglefunc,
        tag=str(tag),
        img=static('plus.gif') if small else static('minus.gif'),
        title_html=title_html,
    )


def visibility_contentdiv(tag: str,
                          contents: str,
                          extra_div_classes: Iterable[str] = None,
                          small: bool = True,
                          as_visibility: bool = True) -> str:
    extra_div_classes = extra_div_classes or []
    div_classes = ["collapsible"] + extra_div_classes
    if as_visibility:
        if small:
            div_classes.append("collapse_invisible")
        else:
            div_classes.append("collapse_visible")
    else:
        if small:
            div_classes.append("collapse_small")
        else:
            div_classes.append("collapse_big")
    return """
<div class="{div_classes}" id="collapsible_{tag}">
    {contents}
</div>
    """.format(
        div_classes=" ".join(div_classes),
        tag=str(tag),
        contents=contents,
    )


def visibility_div_with_divbutton(tag: str,
                                  contents: str,
                                  title_html: str = '',
                                  extra_div_classes: Iterable[str] = None,
                                  small: bool = True) -> str:
    # The HTML pre-hides, rather than using an onload method
    button = visibility_button(tag=tag, small=small,
                               title_html=title_html, as_visibility=True)
    contents = visibility_contentdiv(tag=tag, contents=contents,
                                     extra_div_classes=extra_div_classes,
                                     small=small, as_visibility=True)
    return "<div>" + button + contents + "</div>"


def overflow_div(tag: str,
                 contents: str,
                 extra_div_classes: Iterable[str] = None,
                 small: bool = True) -> str:
    button = visibility_button(tag=tag, small=small,
                               as_visibility=False)
    contentdiv = visibility_contentdiv(tag=tag, contents=contents,
                                       extra_div_classes=extra_div_classes,
                                       small=small, as_visibility=False)
    return """
<div class="expandcollapsewrapper">
    {button}
    {contentdiv}
</div>
    """.format(button=button, contentdiv=contentdiv)


# =============================================================================
# Class to maintain element counters, for use with pages having lots of
# collapsible divs (or other HTML elements requiring individual numbering)
# =============================================================================

class HtmlElementCounter(object):
    def __init__(self, prefix: str = ''):
        self.elementnum = 0
        self.prefix = prefix

    def next(self):
        self.elementnum += 1

    def tag(self):
        return self.prefix + str(self.elementnum)

    def visibility_div_with_divbutton(self,
                                      contents: str,
                                      title_html: str = '',
                                      extra_div_classes: Iterable[str] = None,
                                      visible: bool = True) -> str:
        result = visibility_div_with_divbutton(
            tag=self.tag(),
            contents=contents,
            title_html=title_html,
            extra_div_classes=extra_div_classes,
            small=visible)
        self.next()
        return result

    def visibility_div_spanbutton(self, small: bool = True) -> str:
        return visibility_button(tag=self.tag(), as_visibility=True,
                                 small=small, as_span=True)

    def visibility_div_contentdiv(self,
                                  contents: str,
                                  extra_div_classes: Iterable[str] = None,
                                  small: bool = True) -> str:
        result = visibility_contentdiv(
            tag=self.tag(),
            contents=contents,
            extra_div_classes=extra_div_classes,
            small=small,
            as_visibility=True)
        self.next()
        return result

    def collapsible_div_contentdiv(self,
                                   contents: str,
                                   extra_div_classes: Iterable[str] = None,
                                   small: bool = True) -> str:
        result = visibility_contentdiv(
            tag=self.tag(),
            contents=contents,
            extra_div_classes=extra_div_classes,
            small=small,
            as_visibility=False)
        self.next()
        return result

    def overflow_div(self,
                     contents: str,
                     extradivclasses: Iterable[str] = None,
                     small: bool = True) -> str:
        result = overflow_div(tag=self.tag(),
                              contents=contents,
                              extra_div_classes=extradivclasses,
                              small=small)
        self.next()
        return result


# =============================================================================
# Highlighting of query results
# =============================================================================

HIGHLIGHT_FWD_REF = "Highlight"


def escape_literal_string_for_regex(s: str) -> str:
    r"""
    Escape any regex characters.

    Start with \ -> \\
        ... this should be the first replacement in REGEX_METACHARS.
    """
    for c in REGEX_METACHARS:
        s.replace(c, "\\" + c)
    return s


def get_regex_from_highlights(highlight_list: Iterable[HIGHLIGHT_FWD_REF],
                              at_word_boundaries_only: bool = False) \
        -> typing.re.Pattern:
    elements = []
    wb = r"\b"  # word boundary; escape the slash if not using a raw string
    for hl in highlight_list:
        h = escape_literal_string_for_regex(hl.text)
        if at_word_boundaries_only:
            elements.append(wb + h + wb)
        else:
            elements.append(h)
    regexstring = u"(" + "|".join(elements) + ")"  # group required, to replace
    return re.compile(regexstring, re.IGNORECASE | re.UNICODE)


def highlight_text(x: str, n: int = 0) -> str:
    n %= N_CSS_HIGHLIGHT_CLASSES
    return r'<span class="highlight{n}">{x}</span>'.format(n=n, x=x)


def make_highlight_replacement_regex(n: int = 0) -> str:
    return highlight_text(r"\1", n=n)


def make_result_element(x: Optional[str],
                        element_counter: HtmlElementCounter,
                        highlight_dict: Dict[int,
                                             List[HIGHLIGHT_FWD_REF]] = None,
                        collapse_at_len: int = None,
                        collapse_at_n_lines: int = None,
                        line_length: int = None,
                        keep_existing_newlines: bool = True,
                        collapsed: bool = True,
                        null: str = '<i>NULL</i>') -> str:
    # return escape(repr(x))
    if x is None:
        return null
    highlight_dict = highlight_dict or {}
    x = str(x)
    xlen = len(x)  # before we mess around with it
    # textwrap.wrap will absorb existing newlines
    if keep_existing_newlines:
        input_lines = x.split("\n")
    else:
        input_lines = [x]
    if line_length:
        output_lines = []
        for line in input_lines:
            if line:
                output_lines.extend(textwrap.wrap(line, width=line_length))
            else:  # blank line; textwrap.wrap will swallow it
                output_lines.append('')
    else:
        output_lines = input_lines
    n_lines = len(output_lines)
    # return escape(repr(output_lines))
    output = linebreaksbr(escape("\n".join(output_lines)))
    # return escape(repr(output))
    for n, highlight_list in highlight_dict.items():
        find = get_regex_from_highlights(highlight_list)
        replace = make_highlight_replacement_regex(n)
        output = find.sub(replace, output)
    if ((collapse_at_len and xlen >= collapse_at_len) or
            (collapse_at_n_lines and n_lines >= collapse_at_n_lines)):
        result = element_counter.overflow_div(contents=output,
                                              small=collapsed)
        element_counter.next()
    else:
        result = output
    return result


def pre(x: str = '') -> str:
    return "<pre>{}</pre>".format(x)


# =============================================================================
# SQL formatting
# =============================================================================

SQL_BASE_CSS_CLASS = "sq"  # brief is good
SQL_FORMATTER = HtmlFormatter(cssclass=SQL_BASE_CSS_CLASS)
SQL_LEXER = SqlLexer()


def prettify_sql_html(sql: str,
                      reformat: bool = False,
                      indent_width: int = 4) -> str:
    if reformat:
        sql = sqlparse.format(sql, reindent=True, indent_width=indent_width)
    return highlight(sql, SQL_LEXER, SQL_FORMATTER)


@django_cache_function(timeout=None)
def prettify_sql_css() -> str:
    return SQL_FORMATTER.get_style_defs()


def prettify_sql_and_args(sql: str, args: List[Any] = None,
                          reformat: bool = False,
                          indent_width: int = 4) -> str:
    sql = prettify_sql_html(sql, reformat=reformat, indent_width=indent_width)
    if args:
        formatted_args = "\n".join(textwrap.wrap(repr(args)))
        return sql + "<div>Args:</div><pre>{}</pre>".format(formatted_args)
    else:
        return sql


def make_collapsible_sql_query(x: Optional[str],
                               element_counter: HtmlElementCounter,
                               args: List[Any] = None,
                               collapse_at_len: int = 400,
                               collapse_at_n_lines: int = 5) -> str:
    x = x or ''
    x = str(x)
    xlen = len(x)
    n_lines = len(x.split('\n'))
    formatted = prettify_sql_and_args(x, args, reformat=False)
    # x = linebreaksbr(escape(x))
    if ((collapse_at_len and xlen >= collapse_at_len) or
            (collapse_at_n_lines and n_lines >= collapse_at_n_lines)):
        return element_counter.overflow_div(contents=formatted)
    return formatted



#!/usr/bin/env python
# crate_anon/crateweb/core/middleware.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import reverse
from django.http import HttpResponse, HttpResponseForbidden, HttpRequest
from crate_anon.crateweb.core.utils import is_developer


# =============================================================================
# RestrictAdminMiddleware
# =============================================================================
# See https://djangosnippets.org/snippets/2227/


class RestrictAdminMiddleware(object):
    """
    A middleware that restricts different admin sites depending on user
    privileges.
    """
    @staticmethod
    def process_request(request: HttpRequest) -> HttpResponse:
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "Authentication middleware required. Edit your"
                " MIDDLEWARE_CLASSES setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the RestrictDevAdminMiddleware class.")
        if request.path.startswith(reverse('devadmin:index')):
            if not is_developer(request.user):
                return HttpResponseForbidden(
                    "Non-developers cannot access the devadmin")
        if request.path.startswith(reverse('mgradmin:index')):
            if not request.user.is_superuser:
                return HttpResponseForbidden(
                    "Non-superusers cannot access the mgradmin")
        # Django requires the staff flag for any admin.



#!/usr/bin/env python
# crate_anon/nlp_manager/number.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from typing import Optional


def to_float(s: str) -> Optional[float]:
    if s:
        s = s.replace(',', '')  # comma as thousands separator
        s = s.replace('−', '-')  # Unicode minus
        s = s.replace('–', '-')  # en dash
    try:
        return float(s)
    except (TypeError, ValueError):
        return None


def to_pos_float(s: str) -> Optional[float]:
    try:
        return abs(to_float(s))
    except TypeError:  # to_float() returned None
        return None



#!/usr/bin/env python
# crate_anon/nlp_manager/lanuch_multiprocess_nlp.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

# Launch nlp_manager.py in multiprocess mode.
#
# Previous bash version:
#
# http://stackoverflow.com/questions/356100
# http://stackoverflow.com/questions/1644856
# http://stackoverflow.com/questions/8903239
# http://stackoverflow.com/questions/1951506
# Note: $! is the process ID of last process launched in background
# http://stackoverflow.com/questions/59895
#
# Python version:
#
# http://stackoverflow.com/questions/23611396/python-execute-cat-subprocess-in-parallel  # noqa
# http://stackoverflow.com/questions/320232/ensuring-subprocesses-are-dead-on-exiting-python-program  # noqa
# http://stackoverflow.com/questions/641420/how-should-i-log-while-using-multiprocessing-in-python  # noqa

import argparse
# import atexit
import logging
import multiprocessing
import sys
import time
from typing import List

from crate_anon.common.subproc import (
    check_call_process,
    run_multiple_processes,
)
from crate_anon.common.logsupport import configure_logger_for_colour
from crate_anon.version import VERSION, VERSION_DATE

log = logging.getLogger(__name__)

NLP_MANAGER = 'crate_anon.nlp_manager.nlp_manager'

CPUCOUNT = multiprocessing.cpu_count()


# =============================================================================
# Main
# =============================================================================

def main() -> None:
    version = "Version {} ({})".format(VERSION, VERSION_DATE)
    description = (
        "Runs the CRATE NLP manager in parallel. {}. Note that all arguments "
        "not specified here are passed to the underlying script "
        "(see crate_nlp --help).".format(version))
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "--nlpdef", required=True,
        help="NLP processing name, from the config file")
    parser.add_argument(
        "--nproc", "-n", nargs="?", type=int, default=CPUCOUNT,
        help="Number of processes (default on this "
             "machine: {})".format(CPUCOUNT))
    parser.add_argument(
        '--verbose', '-v', action='store_true',
        help="Be verbose")
    args, unknownargs = parser.parse_known_args()

    loglevel = logging.DEBUG if args.verbose else logging.INFO
    rootlogger = logging.getLogger()
    configure_logger_for_colour(rootlogger, loglevel)

    common_options = (
        ['--nlpdef', args.nlpdef] +
        ["-v"] * (1 if args.verbose else 0) +
        unknownargs
    )

    log.debug("common_options: {}".format(common_options))

    nprocesses_main = args.nproc
    # nprocesses_index = args.nproc

    # -------------------------------------------------------------------------
    # Setup
    # -------------------------------------------------------------------------

    # Start.
    time_start = time.time()

    # -------------------------------------------------------------------------
    # Clean/build the tables. Only run one copy of this!
    # -------------------------------------------------------------------------
    # CALL USING "python -m my.module"; DO NOT run the script as an executable.
    # If you run a Python script as an executable, it gets added to the
    # PYTHONPATH. Then, when your script says "import regex" (meaning the
    # system module), it might import "regex.py" from the same directory (which
    # it wouldn't normally do, because Python 3 uses absolute not relative
    # imports).
    procargs = [
        sys.executable, '-m', NLP_MANAGER,
        '--dropremake',
        '--processcluster', 'STRUCTURE'
    ] + common_options
    check_call_process(procargs)

    # -------------------------------------------------------------------------
    # Now run lots of things simultaneously:
    # -------------------------------------------------------------------------
    # (a) patient tables
    args_list = []  # type: List[List[str]]
    for procnum in range(nprocesses_main):
        procargs = [
            sys.executable, '-m', NLP_MANAGER,
            '--nlp',
            '--processcluster', 'NLP',
            '--nprocesses={}'.format(nprocesses_main),
            '--process={}'.format(procnum)
        ] + common_options
        args_list.append(procargs)
    run_multiple_processes(args_list)  # Wait for them all to finish

    # time_middle = time.time()

    # -------------------------------------------------------------------------
    # We used to index at the end.
    # (Always fastest to index last.)
    # But now we combine index definitions with column definitions in SQLA.
    # -------------------------------------------------------------------------
    # args_list = [
    #     [
    #         sys.executable, '-m', NLP_MANAGER,
    #         '--index',
    #         '--processcluster=INDEX',
    #         '--nprocesses={}'.format(nprocesses_index),
    #         '--process={}'.format(procnum)
    #     ] + common_options for procnum in range(nprocesses_index)
    # ]
    # run_multiple_processes(args_list)  # Wait for them all to finish

    # -------------------------------------------------------------------------
    # Finished.
    # -------------------------------------------------------------------------
    time_end = time.time()

    # main_dur = time_middle - time_start
    # index_dur = time_end - time_middle
    # total_dur = time_end - time_start
    # print("Time taken: main {} s, indexing {} s, total {} s".format(
    #     main_dur, index_dur, total_dur))

    total_dur = time_end - time_start
    print("Time taken: {} s".format(total_dur))


if __name__ == '__main__':
    main()
